"""Monitor the plugin co-ordinator and check if it is Out of Memory, perform heart beats and capture timeouts."""

import contextlib
import datetime
import json
import logging
import multiprocessing
import os
import pathlib
import shutil
import signal
import tempfile
import time
import traceback
import uuid
from typing import Callable

import psutil
import pydantic
from azul_bedrock import models_network as azm

from azul_runner.coordinator import (
    Coordinator,
    CriticalError,
    RecreateException,
    SigTermExitError,
    get_git_version_suffix,
)
from azul_runner.log_setup import AddLoggingQueueListener, LogLevel, setup_logger
from azul_runner.models import JobResult, State, TaskExitCodeEnum, TaskModel

from . import network, settings
from . import plugin as mplugin
from .storage import StorageProxyFile

MAX_ATTEMPTS_TO_KILL_CHILD_PROCESS = 10
LOCAL_RESULT_SUFFIX = "-result"
logger = logging.getLogger(__name__)

result_type_adapter = pydantic.TypeAdapter(dict[str | None, JobResult])


class PluginTimeoutError(RuntimeError):
    """Raised if a child process executing the plugin can't be killed, or there were too many plugin timeouts."""

    pass


class TerminateError(RuntimeError):
    """Raised if a child process exited and requested to be terminated."""

    pass


class NoNetworkResultError(Exception):
    """Error to raise when a heartbeat or OOM error can't be sent to dispatcher because there is no network."""

    def __init__(self, message: str, result: JobResult):
        super().__init__(message, result)
        self.result = result


def kill_child_proc_tree(pid: int, sig=signal.SIGKILL):
    """Kill all child processes via process tree (including grandchildren) with signal KILL."""
    try:
        # If the parent no longer exists you can't kill children.
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        return

    children = parent.children(recursive=True)
    for p in children:
        try:
            p.send_signal(sig)
        except psutil.NoSuchProcess:
            pass


class RunOnceHandler:
    """Handle multiprocessing when only running the subprocess for one iteration."""

    def __init__(
        self,
        event: azm.BinaryEvent,
        local_streams: list[StorageProxyFile] = None,
        cache_dir_path: str = tempfile.gettempdir(),
    ):
        """Handle the communication between a child process and the parent process."""
        self.event = event
        self.local_streams = local_streams
        self.cache_dir_path = cache_dir_path
        self.run_guid = str(uuid.uuid4())

    def cleanup_streams(self):
        """Cleanup any open streams."""
        if not self.local_streams:
            return
        for stream in self.local_streams:
            if not stream.closed:
                stream.close()

    def _get_result_file_path(self) -> str:
        """Get the results file path."""
        return os.path.join(self.cache_dir_path, self.run_guid + LOCAL_RESULT_SUFFIX)

    def _generate_stream_path(self, file_label) -> str:
        """Generate a path for a stream."""
        return os.path.join(self.cache_dir_path, file_label + "-" + self.run_guid)

    def load_job_results_from_temp(self) -> dict[str, JobResult]:
        """Loads results or raises an exception if there are no results.

        NOTE - leaves open file handles for all streams to allow for tests to inspect streams.
        """
        try:
            with open(self._get_result_file_path(), "rb") as read_result:
                results = dict()
                for multiplugin_name, result_value in json.loads(read_result.read()).items():
                    # Account for serialization/deserialization of none
                    if multiplugin_name.lower() == "none":
                        multiplugin_name = None
                    results[multiplugin_name] = JobResult.model_validate(result_value)
        except Exception as e:
            with open(self._get_result_file_path(), "rb") as read_result:
                file_data = read_result.read()
            logger.info(f"Input data was {file_data} with error {e}")
            raise
        # Leave all the files as open file handles.
        for r in results.values():
            for file_label in r.data:
                r.data[file_label] = open(self._generate_stream_path(file_label), "rb")
        # Cleanup result file, otherwise they build up in temp.
        with contextlib.suppress(Exception):
            os.remove(self._get_result_file_path())
        return results

    def _save_job_results_to_temp(self, result_dict: dict[str, JobResult]):
        """Save the job result to the provided temp directory."""
        for result in result_dict.values():
            for label, file_handle in result.data.items():
                file_handle.seek(0)
                with open(self._generate_stream_path(label), "wb") as stream_file:
                    stream_file.write(file_handle.read())
                # Close the file because this is the final read.
                file_handle.close()
                result.data[label] = b""
        json_results = result_type_adapter.dump_json(result_dict, round_trip=True)
        with open(self._get_result_file_path(), "wb") as out_file:
            out_file.write(json_results)

    def start_run_once_coordinator(
        self,
        *,
        plugin: type[mplugin.Plugin],
        config: settings.Settings,
        job_limit: int,  # Ignored but kept for compatibility.
        log_level: LogLevel,
        queue: multiprocessing.Queue,
        logging_queue: multiprocessing.Queue,
    ):
        """Run a plugin once and then write the result to the cache_dir_path."""
        try:
            setup_logger(log_level, logging_queue)
            loop = Coordinator(plugin, config)
            # Note - ignoring multi-plugin for timeouts.
            if queue:
                queue.put(TaskModel(in_event=self.event, start_time_epoch=time.time()), timeout=0.5)
            results = loop.run_once(self.event, self.local_streams)
            self._save_job_results_to_temp(results)
        finally:
            self.cleanup_streams()


def _start_loop_coordinator(
    *,
    plugin: type[mplugin.Plugin],
    config: settings.Settings,
    job_limit: int,
    log_level: LogLevel,
    queue: multiprocessing.Queue,
    logging_queue: multiprocessing.Queue,
):
    """Start the coordinator and plugin and return exit code if it exits."""
    setup_logger(log_level, logging_queue)
    loop = Coordinator(plugin, config)
    pid = os.getpid()
    try:
        loop.run_loop(queue=queue, job_limit=job_limit)
        # clean exit
        logger.info("Plugin has stopped cleanly.")
        exit(TaskExitCodeEnum.COMPLETED)
    except CriticalError:
        logger.info("closing program after critical error.")
        exit(TaskExitCodeEnum.TERMINATE)
    except SigTermExitError:
        logger.info("closing program after SigTerm recieved error.")
        exit(TaskExitCodeEnum.TERMINATE)
    except RecreateException:
        # Kill process tree to remove any bad child processes launched with subprocess module.
        logger.info("Re-creating the plugin.")
        kill_child_proc_tree(pid)
        exit(TaskExitCodeEnum.RECREATE_PLUGIN)
    except Exception:
        logger.error(f"Plugin crashed with generic error {traceback.format_exc()}")
        exit(TaskExitCodeEnum.TERMINATE)


class MonitorTask:
    """Object that has the current states of a running plugin that is being tracked by Monitor.

    Coordinator is the child_process that is being tracked.
    Queue is the multiprocessing queue that is held by the main process and this process.
        This enables the child_process to communicate how many and what jobs it has worked on.

    The other class attributes are used to track heart beating and queuing information.
    """

    def __init__(self, child_process: multiprocessing.Process, queue: multiprocessing.Queue):
        self.child_process = child_process
        self.current_job: TaskModel | None = None
        self.job_sent_heartbeats = 0
        self.job_low_memory_warning_raised = False
        self.timeout_count = 0
        self.queue = queue

    def set_current_job_and_count_completed_jobs(self) -> int:
        """Count the number of completed jobs and set the current_job to None or a task.

        Sets the current_job to a non-None value if the plugin is actively working on a job.
        """
        counted_jobs = 0
        while not self.queue.empty():
            self.current_job = self.queue.get(block=False)
            # Only increase count for main plugin runs, not multiplugins.
            if self.current_job and self.current_job.multi_plugin_name is None:
                counted_jobs += 1
            self.job_sent_heartbeats = 0
            self.job_low_memory_warning_raised = False
        return counted_jobs


class Monitor:
    """Monitor the plugin coordinator and check if it is Out of Memory, perform heart beats and capture timeouts."""

    def __init__(self, plugin_class: type[mplugin.Plugin], config: dict):
        """Init."""
        self._plugin_class = plugin_class
        self._cfg = settings.parse_config(self._plugin_class, config)
        self.mp_ctx = multiprocessing.get_context("forkserver")

        # Setup plugin
        self._recreate_plugin()

        # Calculate time to wait once as it won't change.
        # Default to 1 second because monitor needs to continually clear the Queue between monitor and coordinator
        # or there is a risk the queue becomes full and then the plugin will be stuck until the queue clears.
        self.time_to_wait_between_checks = 1
        if self._cfg.enable_mem_limits:
            self.time_to_wait_between_checks = min(
                self.time_to_wait_between_checks, self._cfg.mem_poll_frequency_milliseconds / 1000
            )

        # Should only be set to true for run once local runs to prevent sending messages to dispatcher.
        self.no_network = False

        # Memory limit setup
        self.max_memusage_bytes = -1
        self.warn_memusage_bytes = -1
        self.error_memusage_bytes = -1
        if self._cfg.enable_mem_limits:
            try:
                max_mem_bytes = -1
                with open(self._cfg.max_mem_file_path, "r") as f:
                    max_mem_bytes = f.read()
                    self.max_memusage_bytes = int(max_mem_bytes)
                    self.warn_memusage_bytes = int(self.max_memusage_bytes * self._cfg.used_mem_warning_frac)
                    self.error_memusage_bytes = int(self.max_memusage_bytes * self._cfg.used_mem_force_exit_frac)

            except ValueError:
                logger.warning(f"Invalid mem usage from file ('{max_mem_bytes}'), disabling mem checks!")
                self._cfg.enable_mem_limits = False
            except Exception:
                if max_mem_bytes == -1:
                    logger.error(f"Could not find file '{self._cfg.max_mem_file_path}' it doesn't exist.")
                else:
                    logger.error(
                        f"Could not cast the value '{max_mem_bytes}' into an integer to set max_memusage_bytes."
                    )
                raise

    def _recreate_plugin(self):
        """Recreate the plugin and networking for monitor."""
        version_suffix = get_git_version_suffix(self._cfg)
        if version_suffix:
            self._cfg.version_suffix = version_suffix
        self._plugin = self._plugin_class(config=self._cfg)
        self._network = network.Network(self._plugin)

    # Tempfile.template should be "tmp" which is the default tempfile prefix when creating files.
    @staticmethod
    def delete_tempfiles(file_prefix: str = tempfile.template):
        """Delete temporary files if any exists, this prevents lost tempfiles building up."""
        try:
            temp_dir = pathlib.Path(tempfile.gettempdir())
            for file in temp_dir.iterdir():
                # All temp files created by a plugin should start with tmp
                if file.is_file() and file.name.lower().startswith(file_prefix.lower()):
                    file.unlink(missing_ok=True)
                    logger.info(f"azul-runner is deleting temporary file {file} for cleanup purposes.")
        except Exception:
            logger.warning(
                f"Unable to clear all the files from temp dir '{tempfile.gettempdir()}'"
                + f" while restarting the plugin with error: {traceback.format_exc()}"
            )

    @staticmethod
    def purge_temp_directory():
        """Delete everything in the temporary directory (needed during plugin recreation requests)."""
        folder = tempfile.gettempdir()
        try:
            for filename in os.listdir(folder):
                file_path = os.path.join(folder, filename)
                # Delete file
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.remove(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
        except Exception:
            logger.warning(f"Failed to fully cleanup temp: {traceback.format_exc()}")

    def _create_and_start_child_process(
        self,
        start_child_process_func: Callable[
            [
                type[mplugin.Plugin],
                settings.Settings,
                int,
                LogLevel,
                multiprocessing.Queue,
                multiprocessing.Queue,
            ],
            None,
        ],
        job_limit: int,
        queue: multiprocessing.Queue,
        logging_queue: multiprocessing.Queue,
    ) -> multiprocessing.Process:
        """Create the child co-ordinator process."""
        # Ensure that if the process was re-created the parent process waits until fetching is done.
        queue.empty()
        queue.put(None, timeout=0.5)
        log_level = logging.getLogger().level
        child_process = self.mp_ctx.Process(
            target=start_child_process_func,
            kwargs=dict(
                plugin=self._plugin_class,
                config=self._cfg,
                job_limit=job_limit,
                log_level=log_level,
                queue=queue,
                logging_queue=logging_queue,
            ),
        )
        child_process.start()

        return child_process

    def _kill_child_processes(self, concurrent_task_list: list[MonitorTask]):
        """Continually kill the child process until it's no longer alive."""
        killing_attempt = 0
        for cur_task in concurrent_task_list:
            while cur_task.child_process.is_alive():
                # Keep trying to kill the child process and all of it's children.
                kill_child_proc_tree(cur_task.child_process.pid)
                cur_task.child_process.kill()
                time.sleep(0.1)
                killing_attempt += 1
                if killing_attempt > MAX_ATTEMPTS_TO_KILL_CHILD_PROCESS:
                    # Failed to kill child process.
                    raise TimeoutError()

    def run_once(
        self,
        event: azm.BinaryEvent,
        local_streams: list[StorageProxyFile] = None,
        cache_dir_path: str = tempfile.tempdir,
        no_network=True,
    ) -> dict[str, JobResult]:
        """Run the plugin once in a subprocess and get the result.

        no_network prevents heartbeats from being sent and errors if there is memory or timeout issues.
        """
        roh = RunOnceHandler(event, local_streams, cache_dir_path)
        # Shouldn't wait anytime between checks in testing.
        self.time_to_wait_between_checks = 0
        self.no_network = no_network
        # Run and get result
        self._run(roh.start_run_once_coordinator)
        return roh.load_job_results_from_temp()

    def run_loop(self, job_limit: int | None = None):
        """Run a plugin in an infinite loop or until the job limit is reached."""
        self._run(_start_loop_coordinator, job_limit)

    def propagate_termination_signal(self, tasks: list[MonitorTask], send_sig: signal.Signals):
        """Setup a function to propagate the termination of this process to the child process."""

        def func_terminate_all_child_processes(*args):
            for t in tasks:
                t.child_process.terminate()

        signal.signal(send_sig, func_terminate_all_child_processes)

    def _run(
        self,
        start_child_process_func: Callable[
            [
                type[mplugin.Plugin],
                settings.Settings,
                int,
                LogLevel,
                multiprocessing.Queue,
                multiprocessing.Queue,
            ],
            None,
        ],
        job_limit=None,
    ):
        """Run a child plugin in a subprocess infinite loop or until the job limit is reached."""
        job_count = 0
        logging_queue = self.mp_ctx.Queue()
        concurrent_task_list: list[MonitorTask] = []
        recreate_plugin_requested = False
        plugin_clean_exit_requested = False
        is_any_job_active = False
        try:
            with AddLoggingQueueListener(logging_queue, logging.getLogger()):
                # Delete temp files out of temp if any exist. This prevents tmp files building up.
                self.delete_tempfiles()
                for _ in range(self._cfg.concurrent_plugin_instances):
                    queue = self.mp_ctx.Queue()
                    child_process = self._create_and_start_child_process(
                        start_child_process_func, job_limit, queue, logging_queue
                    )
                    concurrent_task_list.append(MonitorTask(child_process, queue))
                # Ensure SIGINT and SIGTERM cause the child process to terminate.
                self.propagate_termination_signal(concurrent_task_list, signal.SIGINT)
                self.propagate_termination_signal(concurrent_task_list, signal.SIGTERM)

                while True:
                    # Sleep only if the job_limit isn't set and if the limit is set start waiting once you have
                    # processed that many jobs. (makes testing much faster)
                    if job_limit and job_count > job_limit:
                        logger.info("Exiting after reaching job limit.")
                        return
                    elif job_limit is None or job_count > job_limit:
                        time.sleep(self.time_to_wait_between_checks)
                        pass

                    # Confirm at least one task wants to be recreated and none have any active jobs.
                    if recreate_plugin_requested and not is_any_job_active:
                        self.purge_temp_directory()
                        self._recreate_plugin()
                        # Ensure all child processes were terminated before re-creating them.
                        self._kill_child_processes(concurrent_task_list)
                        for monitor_task in concurrent_task_list:
                            # Ensure the jobs in the child tasks queue are cleared before starting plugin to prevent
                            # Deadlocks.
                            monitor_task.set_current_job_and_count_completed_jobs()
                            # Start child process again
                            monitor_task.child_process = self._create_and_start_child_process(
                                start_child_process_func, job_limit, queue, logging_queue
                            )

                    # Confirm at least one task wants to exit and there are no active tasks.
                    if plugin_clean_exit_requested and not is_any_job_active:
                        logger.info("Closing down successfully completed plugin.")
                        return

                    recreate_plugin_requested = False
                    plugin_clean_exit_requested = False
                    is_any_job_active = False

                    # If the child process has stopped handle that case.
                    for monitor_task in concurrent_task_list:
                        if not child_process.is_alive():
                            if monitor_task.child_process.exitcode == TaskExitCodeEnum.COMPLETED.value:
                                plugin_clean_exit_requested = True
                                continue
                            elif monitor_task.child_process.exitcode == TaskExitCodeEnum.RECREATE_PLUGIN.value:
                                recreate_plugin_requested = True
                                continue
                            elif monitor_task.child_process.exitcode == TaskExitCodeEnum.TERMINATE.value:
                                raise TerminateError("Critical error occurred crashing process.")
                            else:
                                error_message = (
                                    "Unknown exit code from child process code: "
                                    + f"'{monitor_task.child_process.exitcode}' crashing!"
                                )
                                logger.error(error_message)
                                raise TerminateError(error_message)

                        # Get the current job in the queue (also ensures it empties out)\
                        job_count += monitor_task.set_current_job_and_count_completed_jobs()

                        # Still fetching a new job, or haven't got the first one yet (both have current_job of None)
                        if monitor_task.current_job is None:
                            continue

                        # A job is active because it has a current_job that isn't None
                        # and the subprocess hasn't terminated for any reason.
                        is_any_job_active = True

                        # Perform memory checks if enabled
                        if self._cfg.enable_mem_limits:
                            if not self._are_memory_limits_good(monitor_task):
                                # prevent temp file buildup.
                                self._kill_child_processes(concurrent_task_list)
                                self.delete_tempfiles()
                                child_process = self._create_and_start_child_process(
                                    start_child_process_func, job_limit, queue, logging_queue
                                )

                        # Check for timeout and raise heartbeat if required.
                        if not self._is_healthy_heartbeat_and_memory_checks(monitor_task):
                            self._kill_child_processes(concurrent_task_list)
                            # prevent temp file buildup.
                            self.delete_tempfiles()
                            child_process = self._create_and_start_child_process(
                                start_child_process_func, job_limit, queue, logging_queue
                            )

        finally:
            self._kill_child_processes(concurrent_task_list)
            # Close all the queues.
            for c_task in concurrent_task_list:
                c_task.queue.close()
            logging_queue.close()

    def _refine_current_memory_value(self, current_mem_bytes) -> int:
        """Further refine a memory value by subtracting the inactive file from the total amount of memory in use.

        This prevents premature exiting of an application.
        """
        cur_inactive_file_memory = 0
        with open(self._cfg.cur_mem_summary_file_path, "r") as cur_mem_summary_file:
            try:
                for line in cur_mem_summary_file.readlines():
                    if line.startswith("inactive_file"):
                        line = line.removeprefix("inactive_file")
                        cur_inactive_file_memory = int(line.strip())
                        break
            except ValueError:
                logger.warning("Couldn't read memory summary, memory readings may be inaccurate.")
        return current_mem_bytes - cur_inactive_file_memory

    def _are_memory_limits_good(self, monitor_task: MonitorTask):
        """Check if the memory limit has been reached and kill the child process if required.

        Return True if the child process is healthy and False if it should be terminated.
        """
        # Assuming that if max memory exists and was an int cur_mem will be to (save processing time).
        with open(self._cfg.cur_mem_file_path, "r") as cur_mem_file:
            # Account for case when memory file is empty or unreadable.
            try:
                cur_mem_bytes = int(cur_mem_file.read())
            except ValueError as e:
                logger.warning(f"Couldn't read memory file with error {e}")
                return True

        # Memory is at error levels so perform additional refinement of the memory value to verify it will cause OOM.
        if cur_mem_bytes >= self.error_memusage_bytes:
            print(f"ATTEMPTING TO REFINE MEMORY USAGE! {cur_mem_bytes}")
            cur_mem_bytes = self._refine_current_memory_value(cur_mem_bytes)
            print(f"ATTEMPTING TO REFINE AFTER VALUE! {cur_mem_bytes}")

        if cur_mem_bytes >= self.error_memusage_bytes:
            multiplugin_text = self._plugin.NAME
            if monitor_task.current_job.multi_plugin_name:
                multiplugin_text += f"-{monitor_task.current_job.multi_plugin_name}"
            # Ran out of memory kill child process.
            message = (
                f"Plugin {multiplugin_text} failed to complete "
                + f"job '{monitor_task.current_job.in_event.entity.sha256}' because it ran out of memory, "
                + f"memory limit is {pydantic.ByteSize(self.max_memusage_bytes).human_readable()} and memory "
                + f"usage was {pydantic.ByteSize(cur_mem_bytes).human_readable()} which is "
                + f"{(cur_mem_bytes / self.max_memusage_bytes) * 100:.1f}% memory usage."
            )
            logger.error(message)

            result = JobResult(
                state=State(
                    State.Label.ERROR_OOM,
                    failure_name="Out of Memory",
                    message=message,
                ),
            )
            self._post_status(monitor_task, result)
            return False
        elif not monitor_task.job_low_memory_warning_raised and cur_mem_bytes >= self.warn_memusage_bytes:
            multiplugin_text = self._plugin.NAME
            if monitor_task.current_job.multi_plugin_name:
                multiplugin_text += f"-{monitor_task.current_job.multi_plugin_name}"
            logger.warning(
                f"The job '{monitor_task.current_job.in_event.entity.sha256}' is nearly out of memory it is at "
                + f"{(cur_mem_bytes / self.max_memusage_bytes) * 100:.1f}% memory usage for the plugin "
                + f"{multiplugin_text}."
            )
            monitor_task.job_low_memory_warning_raised = True
        return True

    def _is_healthy_heartbeat_and_memory_checks(self, monitor_task: MonitorTask) -> bool:
        """Check if heartbeats should be created and verify.

        Return True if the child process is healthy and False if it should be terminated.
        """
        # Thread is still running - Raise heartbeat and check if plugin should be forced to time out.
        runtime = time.time() - monitor_task.current_job.start_time_epoch

        if self._cfg.run_timeout and runtime > self._cfg.run_timeout:
            # Kill the plugin as it's run out of time.
            logger.warning(
                f"Execution of {self._plugin.NAME} timed out after {self._cfg.run_timeout} seconds "
                f"for job with id '{monitor_task.current_job.in_event.entity.sha256}'"
            )
            state = State(
                State.Label.ERROR_TIMEOUT,
                failure_name="PluginTimeout Error",
                message=f"{self._plugin.NAME} timed out on job '{monitor_task.current_job.in_event.entity.sha256}'",
            )
            result = JobResult(
                state=state,
                date_start=datetime.datetime.fromtimestamp(monitor_task.current_job.start_time_epoch),
                date_end=datetime.datetime.now(datetime.timezone.utc),
                runtime=int(runtime),
            )
            self._post_status(monitor_task, result)
            monitor_task.timeout_count += 1
            if monitor_task.timeout_count >= self._cfg.max_timeouts_before_exit:
                msg = f"Exiting due to hitting limit of {self._cfg.max_timeouts_before_exit} timeouts"
                logger.error(msg)
                raise PluginTimeoutError(msg)
            return False
        # Only check for heartbeats every heartbeat internal.
        elif (runtime // self._cfg.heartbeat_interval) > monitor_task.job_sent_heartbeats:
            # Send heartbeat.
            monitor_task.job_sent_heartbeats += 1
            result = JobResult(
                state=State(State.Label.HEARTBEAT),
                date_start=datetime.datetime.fromtimestamp(monitor_task.current_job.start_time_epoch),
                date_end=None,
                runtime=int(runtime),
            )
            if self.no_network:
                # Ignore heartbeats if there is no network connection.
                logger.info("Not sending heartbeat as networking is disabled.")
                return True
            self._post_status(monitor_task, result)

        return True

    def _post_status(self, monitor_task: MonitorTask, result: JobResult):
        """Post a status message to dispatcher."""
        if self.no_network:
            raise NoNetworkResultError(
                f"Attempting to post a {result.state} event and networking isn't enabled.", result
            )
        self._network.ack_job(monitor_task.current_job.in_event, result, monitor_task.current_job.multi_plugin_name)



"""Execute Plugin with Job and return JobResult instance."""

import datetime
import json
import logging
import traceback

from azul_bedrock import models_network as azm

from . import models
from . import plugin as mplugin
from .models import FeatureValue, JobResult, State
from .storage import ProxyFileNotFoundError, StorageError

logger = logging.getLogger(__name__)


# true feature limit is 10_000, but reserve some for internal use
MAX_FEATURE_VALUES = 9_500


class ResultError(AssertionError):
    """Used to signal an invalid plugin output in the 'results' dict."""

    pass


def run_plugin_with_job(plugin: mplugin.Plugin, job: models.Job, multiplugin: str) -> JobResult:
    """Execute plugin and handle bad output."""
    run_start = datetime.datetime.now(datetime.timezone.utc)
    # cleanup from last run
    plugin.reset(job)

    # execute and validate
    mp = plugin.get_multiplugin(multiplugin)
    try:
        # filter job if plugin requires data
        if plugin.cfg.assume_streams_available and not job.get_all_data():
            state = State(State.Label.OPT_OUT, message="plugin requires data but none exists")
        else:
            logger.debug(f"Running job with '{multiplugin=}'")
            # Execute the plugin with the provided job data.
            response = mp.callback(job)
            state = response
            # normalise status
            if not state:
                state = State(State.Label.COMPLETED)
            elif isinstance(state, State.Label):
                state = State(state)
            elif not isinstance(state, State):
                raise ResultError(f"Returned data cannot be processed as a State instance: {state}")

        # Validate output format
        result = JobResult(state=state)
        # Catch anything that goes wrong while processing output, and report it as a failed run
        if result.state.label in azm.StatusEnumSuccess:
            input_sids = {s.file_info.sha256 for s in job.get_all_data() or []}
            new_hashes = {y.hash for x in plugin.events for y in x.data}
            all_hashes = input_sids.union(new_hashes).union({None})
            feat_types = set()
            warnings = []
            for event in plugin.events:
                if event.parent_sha256 not in all_hashes:
                    raise ResultError(f"results for non-existent stream: {event.parent_sha256}")
                warning_strings = _process_features(plugin, event.features)
                warnings.extend(warning_strings)

                feat_types.update([x for x in plugin.FEATURES if x.name in event.features])
                if event.info:
                    try:
                        json.dumps(event.info)
                    except TypeError as e:
                        raise ResultError(f"info was not JSON-serialisable: {event.info}") from e

                # sort event contents (including feature values)
                event.sort()

            result.feature_types = sorted(feat_types)
            result.events = [x for x in plugin.events if x.keep()]
            result.data = plugin.data

            if len(warnings) > 0:
                result.state = State(
                    label=State.Label.COMPLETED_WITH_ERRORS,
                    message="Partial completion occurred with the following errors: " + "\n".join(warnings),
                )
            # Automatically set the label to COMPLETED_EMPTY if there are no features or data.
            elif (
                result.state.label == State.Label.COMPLETED
                and len(feat_types) == 0
                and len(result.data) == 0
                and len(event.info) == 0
            ):
                result.state.label = State.Label.COMPLETED_EMPTY

    except ProxyFileNotFoundError:
        if plugin.cfg.assume_streams_available:
            result = JobResult(
                state=State(
                    label=State.Label.OPT_OUT,
                    message="Plugin requires binary content but none exists",
                )
            )
        else:
            result = JobResult(
                state=State(
                    label=State.Label.ERROR_EXCEPTION,
                    failure_name="Plugin failed to handle FileNotFoundError",
                    message=traceback.format_exc(),
                )
            )
    except StorageError:
        # Handled differently from regular exception, as it indicates a problem in the system
        #  rather than an error in the plugin
        # FUTURE Consider handling the case where one stream fails but other streams succeed
        logger.error("Encountered StorageError for entity %s:\n%s" % (job.id, traceback.format_exc()))
        result = JobResult(
            state=State(
                label=State.Label.ERROR_NETWORK,
                failure_name="Storage Failure",
                message=traceback.format_exc(),
            )
        )

    except (ResultError, models.ModelError):
        result = JobResult(
            state=State(
                label=State.Label.ERROR_OUTPUT,
                failure_name="Invalid Plugin Output",
                message=traceback.format_exc(),
            )
        )

    except Exception as e:
        logger.error("%s for %s(%s)\n%s" % (e.__class__.__name__, plugin.NAME, job.id, str(e)))
        result = JobResult(
            state=State(
                label=State.Label.ERROR_EXCEPTION,
                failure_name=e.__class__.__name__,
                message=traceback.format_exc(),
            )
        )

    # reset position of reads for data
    for store in job.get_all_data():
        # will fail on closed files
        store.seek(0)

    # set timing information
    result.date_start = run_start
    now = datetime.datetime.now(datetime.timezone.utc)
    result.date_end = now
    result.runtime = int((now - run_start).total_seconds())

    return result


def _process_features(plugin: mplugin.Plugin, feature_values: dict[str, list[FeatureValue]]) -> list[str]:
    """Verify and normalise feature values produced by a plugin.

    Returns, recoverable warning messages in a list.
    """
    warnings = []
    # verify only registered features were set
    bad_features = set(feature_values.keys()).difference(set(x.name for x in plugin.FEATURES))
    if bad_features:
        raise ResultError(f"Plugin tried to set undeclared features: {sorted(bad_features)}")

    total_values = 0  # total feature values
    feature_counts = []  # list of (feature name, feature value count)
    for feature in plugin.FEATURES:
        if feature.name not in feature_values:
            # feature was not produced
            continue

        # remove duplicate feature values
        feature_values[feature.name] = sorted(set(feature_values[feature.name]))
        num_values = len(feature_values[feature.name])

        # check not too many values for a single feature
        if num_values > plugin.cfg.max_values_per_feature:
            feature_values[feature.name] = feature_values[feature.name][: plugin.cfg.max_values_per_feature]
            warnings.append(
                f"too many values for feature {feature.name} ({num_values}) capping returned values to "
                + f"values at {len(feature_values[feature.name])}"
            )
            num_values = len(feature_values[feature.name])

        # check all values have valid type and valid length
        for v in feature_values[feature.name]:
            if not isinstance(v.value, feature.typeref):
                raise ResultError(
                    "Plugin returned a value with incorrect type "
                    f"({feature.name} should be {feature.typeref}, not {type(v.value)})"
                )
            if isinstance(v.value, (str, bytes)) and len(v.value) > plugin.cfg.max_value_length:
                raise ResultError(f"feature {feature.name} has value that is too long ({v.value[:100]}...)")

        # track problematic numbers of feature values
        total_values += num_values
        feature_counts.append((feature.name, num_values))

    # Remove feature values until enough have been removed to put the total number of features below the limit.
    if total_values > MAX_FEATURE_VALUES:
        num_values_breaking = total_values
        # sort feature counts highest to lowest
        feature_counts.sort(key=lambda x: (x[1], x[0]), reverse=True)
        warnings.append(
            f"too many values for plugin ({num_values_breaking}) only returning first ({MAX_FEATURE_VALUES} values)"
        )
        # blindly drop values from the most common features until below the feature limit.
        for feature_name, num_values in feature_counts:
            num_values_drop = min(total_values - MAX_FEATURE_VALUES, num_values)
            # fix total values count
            total_values -= num_values_drop
            warnings.append(f"dropping {num_values_drop}/{num_values} values from {feature_name}")
            # cull values
            feature_values[feature_name] = feature_values[feature_name][: num_values - num_values_drop]
            if total_values <= MAX_FEATURE_VALUES:
                # reached the limit so no further culling required
                break

    return warnings



"""Higher level interactions with Azul Dispatcher vs dispatcher.py."""

import logging
import time
import traceback
import typing

from azul_bedrock import dispatcher, exceptions_bedrock
from azul_bedrock import models_network as azm

from . import network_transform, plugin
from .models import JobResult, State

logger = logging.getLogger(__name__)

# by default will retry getting events if it received corrupted events
CONTINUE_ON_RECV_CORRUPT_EVENT = True


class Network:
    """Handle plugin communications with dispatcher."""

    def __init__(self, p: plugin.Plugin):
        self.plugin = p

        # Cache of metadata for content that has been posted to the server already.
        # Entries are maintained in order by recent usage, so LRU entries can be discarded when the cache limit is
        # reached. Old entries are only dropped on completion of the current job, never during a run.
        self._cached_file_data: dict[tuple[str, str, str], azm.Datastream] = {}

        self.api = dispatcher.DispatcherAPI(
            events_url=self.plugin.cfg.events_url,
            data_url=self.plugin.cfg.data_url,
            retry_count=self.plugin.cfg.request_retry_count,
            timeout=self.plugin.cfg.request_timeout,
            author_name=self.plugin.NAME,
            author_version=self.plugin.VERSION,
            deployment_key=self.plugin.cfg.deployment_key or self.plugin.NAME,
        )

    def _clear_file_metadata(self):
        """Clear metadata cache between runs."""
        self._cached_file_data = {}

    def post_registrations(self):
        """Post all registration events to dispatcher."""
        for row in network_transform.get_registrations(self.plugin):
            logger.info(f"Registering plugin {row.author.name}-{row.author.version if row.author.version else ''}")
            self.api.submit_events([row], model=azm.ModelType.Plugin)

    def fetch_job(self) -> azm.BinaryEvent:
        """Fetches next job from the queue."""

        def fmt_dict_filters(filt: dict[str, list[str]]) -> list[str]:
            """Merge these filters to network format.

            Potentially should be cached as otherwise same work performed every query.
            """
            ret = []
            for k, v in filt.items():
                ret.append(f"{k},{','.join(v)}")
            return ret

        if not self.plugin.cfg.events_url:
            raise ValueError("Cannot fetch jobs when events_url is None")
        while True:
            try:
                info, events = self.api.get_binary_events(
                    count=1,  # only retrieve 1 event, as if plugin crashes these will not reprocess
                    is_task=True,
                    deadline=10,  # dispatcher has up to 10 seconds to retrieve/filter events
                    # filters
                    require_expedite=self.plugin.cfg.require_expedite,
                    require_live=self.plugin.cfg.require_live,
                    require_historic=self.plugin.cfg.require_historic,
                    require_under_content_size=self.plugin.cfg.filter_max_content_size,
                    require_over_content_size=self.plugin.cfg.filter_min_content_size,
                    require_actions=self.plugin.cfg.filter_allow_event_types,
                    deny_self=self.plugin.cfg.filter_self,
                    require_streams=fmt_dict_filters(self.plugin.cfg.filter_data_types),
                )
            except dispatcher.BadResponseException as e:
                logger.warning(f"Failed to decode event from server ({e.__class__.__name__}):\n{str(e)}")
                logger.warning(f"Failed content:\n{e.content}")
                logger.warning(f"Error:\n{traceback.format_exc()}")
                # immediately repoll for new events
                if CONTINUE_ON_RECV_CORRUPT_EVENT:
                    continue
                raise e

            if info.filtered:
                logger.info(f"{info.filtered} uninteresting events filtered by dispatcher")
                if not info.fetched:
                    # immediately repoll for new events
                    continue
            if not events:
                logger.info("No events fetched or filtered by dispatcher; backing off for 10 seconds")
                time.sleep(10)
                continue
            if len(events) > 1:
                raise ValueError(f"{len(events)} events fetched by dispatcher, only 1 allowed")

            event = events[0]
            return event

    def _gen_status(self, src: azm.BinaryEvent, result: JobResult, multiplugin: str | None = None) -> azm.StatusEvent:
        # transform hash-data dictionary to contain labels as well
        labelled_data = {}
        for hash, data in result.data.items():
            labelled_data.setdefault(hash, ([], data))
        for ev in result.events:
            for data in ev.data:
                labelled_data[data.hash][0].append(data.label)
        # post data to dispatcher and receive metadata
        hash_to_meta = self._post_data(src.source.name, labelled_data)
        author_tmpl = network_transform.gen_author(self.plugin, self.plugin.get_multiplugin(multiplugin))
        status_event = network_transform.gen_processing_events(self.plugin, hash_to_meta, author_tmpl, src, result)

        return status_event

    def ack_job(self, src: azm.BinaryEvent, result: JobResult, multiplugin: str | None = None) -> None:
        """Send the result of the job to the dispatcher."""
        try:
            status_event = self._gen_status(src, result, multiplugin)
        except Exception as e:
            # try to capture error
            status_event = self._gen_status(
                src,
                JobResult(
                    state=State(
                        State.Label.ERROR_OUTPUT,
                        failure_name=e.__class__.__name__,
                        message=traceback.format_exc(),
                    ),
                    date_start=result.date_start,
                    date_end=result.date_end,
                    runtime=result.runtime,
                ),
            )

        # use logging lib to format string for performance
        logger.debug("Posting status event: %s", status_event)
        # Post result
        try:
            self.api.submit_events([status_event], model=azm.ModelType.Status)
        except exceptions_bedrock.NetworkDataException as e:
            # try to capture bad output from plugin
            # rerunning the plugin should produce the same result
            # so no need to capture detailed info
            status_event = self._gen_status(
                src,
                JobResult(
                    state=State(State.Label.ERROR_OUTPUT, message=e.args[0]),
                    date_start=result.date_start,
                    date_end=result.date_end,
                    runtime=result.runtime,
                ),
            )
            self.api.submit_events([status_event], model=azm.ModelType.Status)

        # clear file metadata after events have been submitted
        self._clear_file_metadata()

    def _post_data(
        self, source: str, data: dict[str, tuple[list[azm.DataLabel], typing.BinaryIO]]
    ) -> dict[str, azm.Datastream]:
        """Posts data and receives any calculated metadata.

        The server wants us to set metadata on the data streams in downstream jobs/events.
        After completion, self.posted_data should contain ContentEntry entries for each data stream in `data`.
        """
        ret: dict[str, azm.Datastream] = {}
        if not self.plugin.cfg.data_url or not data:
            return ret
        # post all data
        for data_hash, (labels, binary) in data.items():
            for label in labels:
                # attempt to load from cache
                if (source, label, data_hash) in self._cached_file_data:
                    continue

                logger.info("Posting data to server for %s %s %s" % (source, label, data_hash))
                file_info = self.api.submit_binary(source=source, label=label, data=binary)
                file_info.label = label
                # We'll take the returned APIContentEntry as-is, but later replace the 'label' value in _ack_job
                ret[data_hash] = self._cached_file_data[(source, label, data_hash)] = file_info
        return ret

    def get_cached_file(self, source: str, label: azm.DataLabel, data_hash: str) -> azm.Datastream | None:
        """Get cached file info or none if it doesn't exist."""
        return self._cached_file_data.get((source, label, data_hash))



from __future__ import annotations
from dataclasses import dataclass
from typing import Any

from azul_bedrock import models_network as azm
from pydantic import TypeAdapter
from multiprocessing import shared_memory
from azul_runner import (
    FV,
    Event,
    EventData,
    EventParent,
    Feature,
    FeatureType,
    FeatureValue,
    JobResult,
    State,
    TestPlugin,
)
from azul_runner.models import Job
from tests import plugin_support as sup


SHARED_MEM_NAME = "test_plugin_execution_shared_mem"
TEST_SETTING_KEY = "test_setting"


@dataclass
class PluginExecutionSharableInput:
    feature_values: list[tuple[str, Any]]


class PluginExecutionPluginSharedMem(sup.DummyPluginDefaultSharedMem):
    def get_shared_memory_name(self):
        return SHARED_MEM_NAME

    def execute(self, job):
        shared_mem_val: PluginExecutionSharableInput = self._load_from_shared_memory()
        for fv in shared_mem_val.feature_values:
            self.add_feature_values(fv[0], fv[1])


class TestPluginExecutionWrapper(sup.TestPluginSharedMem):
    """
    Tests the handling of plugin execution by Plugin._exec_wrapper, using the sup.DummyPlugin class and TestPlugin template.
    """

    PLUGIN_TO_TEST = sup.DummyPlugin

    @classmethod
    def setUpClass(cls) -> None:
        sup.cleanup_shared_memory(SHARED_MEM_NAME)
        cls.shared_mem = shared_memory.SharedMemory(name=SHARED_MEM_NAME, create=True, size=10240)
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.shared_mem.close()
        cls.shared_mem.unlink()
        return super().tearDownClass()

    def test_basic_run(self):
        """The wrapper should convert the empty dict return to a valid status and empty features and children."""
        result = self.do_execution()
        self.assertEqual(result.state, State(label=State.Label.COMPLETED_EMPTY))

    class DPNothingOnFailure(sup.DummyPlugin):
        def execute(self, job):
            self._add_child("foo", {})
            self.add_feature_values("example_string", ["should not be returned to caller"])
            self.add_info({"key": "value"})
            return State.Label.OPT_OUT

    def test_nothing_on_failure(self):
        """Check that any features or children are deleted when returning a non-completed status"""
        result = self.do_execution(plugin_class=self.DPNothingOnFailure)
        self.assertEqual(result.state, State(State.Label.OPT_OUT))
        self.assertEqual(result.main, None)
        self.assertEqual(len(result.events), 0)

    class DPExceptionHandling(sup.DummyPlugin):
        def execute(self, job):
            raise RuntimeError("foo")

    def test_exception_handling(self):
        """Check for correct result if the plugin raises an exception"""
        result = self.do_execution(plugin_class=self.DPExceptionHandling)
        m: str = result.state.message
        self.assertRegex(m, ".*\nRuntimeError: foo\n")
        self.assertEqual(result.state, State(State.Label.ERROR_EXCEPTION, "RuntimeError", m))

    class DPReturnNondict(sup.DummyPlugin):
        def execute(self, job):
            return [1, 2, 3, 4]

    def test_return_nondict(self):
        """Verifies error result when the plugin doesn't return a dict"""
        result = self.do_execution(plugin_class=self.DPReturnNondict)
        m: str = result.state.message
        self.assertRegex(m, "Returned data cannot be processed as a State instance")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

    class DPReturnBadStatus(sup.DummyPlugin):
        def execute(self, job):
            return "Not a State"

    def test_return_bad_status(self):
        """
        Verifies error result when the plugin returns result['status'] != State(...).
        Don't check further State verification, as this is tested in azul_runner.structures' unit tests.
        """
        result = self.do_execution(plugin_class=self.DPReturnBadStatus)
        m: str = result.state.message
        self.assertRegex(m, "Returned data cannot be processed as a State instance")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

    class DPReturnBareResult1(sup.DummyPlugin):
        def execute(self, job):
            return State.Label.OPT_OUT

    class DPReturnBareResult2(sup.DummyPlugin):
        def execute(self, job):
            return State(State.Label.ERROR_INPUT, "Plugin had an error")

    def test_return_bare_result(self):
        """
        Verifies that runner wraps a bare State / State.Label return value into a dict.
        """
        result = self.do_execution(plugin_class=self.DPReturnBareResult1)
        self.assertEqual(result.state, State(State.Label.OPT_OUT))

        result = self.do_execution(plugin_class=self.DPReturnBareResult2)
        self.assertEqual(result.state, State(State.Label.ERROR_INPUT, "Plugin had an error"))

    def test_return_feature_nonexistent(self):
        """Tests a plugin returning a feature that's not in FEATURES"""
        # 'Primary' feature test
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("nonexistent_feature", [1])]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        m: str = result.state.message
        self.assertRegex(m, "Plugin tried to set undeclared features: \\['nonexistent_feature'\\]\n")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

        # Multiple invalid features
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(
                feature_values=[
                    ("example_string", ["String value"]),
                    ("nonexistent_subfeature", ["String value", "49"]),
                    ("nonexistent_other", [b"something", b"another"]),
                ]
            ),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        m: str = result.state.message
        self.assertRegex(
            m,
            "Plugin tried to set undeclared features: \\['nonexistent_other', 'nonexistent_subfeature'\\]\n",
        )
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

    def test_return_feature_wrapping(self):
        """
        Tests that returned values get wrapped in a FeatureValue instance, while leaving returned FeatureValues alone.
        """

        # Check feature is wrapped and accepted if correct type
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(
                feature_values=[("example_string", ["String but not a FeatureValue instance"])]
            ),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={"example_string": [FV("String but not a FeatureValue instance")]},
                    )
                ],
            ),
        )

        # Check feature is accepted if returned already wrapped in FeatureValue
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_string", [FeatureValue("String")])]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_string": [FV("String")]})],
            ),
        )

    def test_return_str_for_specials(self):
        """Tests that returning a simple str value for Filepath or Uri features is correctly handled."""

        # Check str is accepted for Filepath and Uri

        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(
                feature_values=[
                    ("example_path", ["~/not/a/Filepath/instance"]),
                    ("example_uri", ["http://foo/", "https://bar/"]),
                ]
            ),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={
                            "example_path": [FV("~/not/a/Filepath/instance")],
                            "example_uri": [FV("http://foo/"), FV("https://bar/")],
                        },
                    )
                ],
            ),
        )

    def test_return_feature_wrong_type(self):
        """Tests a plugin returning a feature whose data type doesn't match the type specified in FEATURES."""

        # 'Primary' feature test
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_int", ["String value"])]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        m: str = result.state.message
        self.assertRegex(
            m,
            "Plugin returned a value with incorrect type "
            "\\(example_int should be <class 'int'>, not <class 'str'>\\)\n",
        )
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

    def test_return_feature_none(self):
        """Tests a plugin fails instantly when adding None feature value."""

        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_string", None)]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_string": []})],
            ),
        )

        # check falsy value is ok
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_string", "")]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_string": [FV("")]})],
            ),
        )

    def test_return_feature_bare_value(self):
        """Tests that _exec_wrapper accepts single bare values of the correct type and converts them to lists"""

        # Simple case
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_int", 300)]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_int": [FV(300)]})],
            ),
        )

        # Simple case inside a FeatureValue
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_int", FeatureValue(300))]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_int": [FV(300)]})],
            ),
        )

        # Bare strings should be accepted for Filepath (and Uri, etc)
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_path", r"c:\foo\bar.txt")]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={"example_path": [FV("c:\\foo\\bar.txt")]},
                    )
                ],
            ),
        )

        # Check that incorrect type is rejected/not autoconverted
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_int", 5.0)]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        m: str = result.state.message
        self.assertRegex(m, "Plugin returned a value with incorrect type")
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

        # Check that an unspecified type accepts any valid feature type
        for testval in sup.VALID_FEATURE_EXAMPLES:
            self._set_shared_memory(
                self.shared_mem,
                PluginExecutionSharableInput(feature_values=[("example_unspec", str(testval))]),
            )
            result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
            self.assertEqual(result.state, State())
            self.assertEqual(
                TypeAdapter(list[Event]).dump_python(result.events),
                TypeAdapter(list[Event]).dump_python(
                    [
                        Event(
                            sha256="test_entity",
                            features={"example_unspec": {FeatureValue(str(testval))}},
                        )
                    ]
                ),
            )

        # Check that unspec feature rejects an invalid feature type
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_unspec", AssertionError)]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        m: str = result.state.message
        self.assertRegex(
            m,
            "Input should be a valid",
        )
        self.assertEqual(result.state, State(State.Label.ERROR_EXCEPTION, "ValidationError", m))

        # Check that a FeatureValue of incorrect type is wrapped but rejected
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_string", FeatureValue(5))]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        m: str = result.state.message
        self.assertRegex(
            m,
            "Plugin returned a value with incorrect type \\(example_string should be <class "
            "'str'>, not <class 'int'>\\)",
        )
        self.assertEqual(result.state, State(State.Label.ERROR_OUTPUT, "Invalid Plugin Output", m))

    class DPReturnDuplicateValues(sup.DummyPlugin):
        def execute(self, job):
            self.add_feature_values(
                "example_string",
                [
                    FeatureValue("foo"),
                    FeatureValue("foo", label="bar"),
                ],
            )
            self.add_feature_values(
                "example_int",
                [
                    FeatureValue(5, label="baz"),
                    FeatureValue(5, label="bar"),
                    FeatureValue(5, label="bam"),
                ],
            )

    def test_return_duplicate_values(self):
        """Tests that a plugin returning multiple copies of the same value no longer causes an error"""

        # Simple list
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_int", [1, 2, 3, 2])]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_int": [FV(1), FV(2), FV(3)]})],
            ),
        )

        # Duplicate value is a FeatureValue
        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_int", [1, 2, 3, FeatureValue(2)])]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_int": [FV(1), FV(2), FV(3)]})],
            ),
        )

        # Labelled feat should be merged with standalone feat for a single result value
        result = self.do_execution(plugin_class=self.DPReturnDuplicateValues)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={
                            "example_int": [FV(5, label="bam"), FV(5, label="bar"), FV(5, label="baz")],
                            "example_string": [FV("foo"), FV("foo", label="bar")],
                        },
                    )
                ],
            ),
        )

    def test_return_invalid_labels(self):
        """Tests the error checking when returned labels are wrong type"""
        # Check FeatureValue refuses non-str labels
        self.assertRaises(
            TypeError,
            FeatureValue,
            5,
            b"not a str",
        )

    class DPReturnLabelledFVs(sup.DummyPlugin):
        def execute(self, job):
            self.add_feature_values(
                "example_string",
                [
                    FeatureValue("a string that has labels", label="label1 set in FV object"),
                    FeatureValue("a string that has labels"),
                ],
            )

    def test_return_labelled_featurevalues(self):
        """Tests that plugins returning FeatureValue fields with labels works correctly"""
        result = self.do_execution(plugin_class=self.DPReturnLabelledFVs)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={
                            "example_string": [
                                FV("a string that has labels"),
                                FV("a string that has labels", label="label1 set in FV object"),
                            ]
                        },
                    )
                ],
            ),
        )

    class DPCompletedWithErrorPluginReturn(sup.DummyPlugin):
        def execute(self, job):
            self.add_feature_values("example_int", [1])
            return State(State.Label.COMPLETED_WITH_ERRORS, message="Plugin had an error but still has useful data")

    def test_completed_with_error_plugin_return(self):
        """Tests that returning a simple str value for Filepath or Uri features is correctly handled."""
        # Check str is accepted for Filepath and Uri
        result = self.do_execution(plugin_class=self.DPCompletedWithErrorPluginReturn)
        self.assertJobResult(
            result,
            JobResult(
                state=State(
                    State.Label.COMPLETED_WITH_ERRORS, message="Plugin had an error but still has useful data"
                ),
                events=[
                    Event(
                        sha256="test_entity",
                        features={
                            "example_int": [FV(1)],
                        },
                    )
                ],
            ),
        )

    def test_feature_value_limit(self):
        """Tests that value counts are properly limited"""

        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_int", [9, 8, 7, 6, 5, 4, 3, 2, 1])]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem, config={"max_values_per_feature": 5})
        self.assertEqual(result.state.label, State.Label.COMPLETED_WITH_ERRORS)
        self.assertIn("too many values for feature example_int (9)", result.state.message)
        print(result.events[0])
        self.assertEqual(len(result.events[0].features["example_int"]), 5)

    def test_feature_value_length_limit(self):
        """Tests that value counts are properly limited"""

        self._set_shared_memory(
            self.shared_mem,
            PluginExecutionSharableInput(feature_values=[("example_string", "wow such a long string")]),
        )
        result = self.do_execution(plugin_class=PluginExecutionPluginSharedMem, config={"max_value_length": 5})
        self.assertEqual(result.state.label, State.Label.ERROR_OUTPUT)
        self.assertEqual(result.state.failure_name, "Invalid Plugin Output")
        self.assertIn(
            "feature example_string has value that is too long (wow such a long string...)", result.state.message
        )

    class DPFeatureValueLimitTotal(sup.DummyPlugin):
        def execute(self, job):
            self.add_feature_values("example_int", list(x for x in range(10010)))

    def test_feature_value_limit_total(self):
        """Tests that value counts are properly limited"""

        result = self.do_execution(
            plugin_class=self.DPFeatureValueLimitTotal, config={"max_values_per_feature": 100000}
        )
        self.assertEqual(result.state.label, State.Label.COMPLETED_WITH_ERRORS)
        self.assertIn("too many values for plugin (10010)", result.state.message)
        self.assertEqual(len(result.events[0].features["example_int"]), 9500)

    class DPFeatureValueLimitTotal2(sup.DummyPlugin):
        def execute(self, job):
            self.add_feature_values("example_path", list(f"{x}" for x in range(10010)))
            self.add_feature_values("example_int", list(x for x in range(10008)))
            self.add_feature_values("example_string", list(f"{x}" for x in range(10009)))

    def test_feature_value_limit_total2(self):
        """Tests that value counts are properly limited"""
        result = self.do_execution(
            plugin_class=self.DPFeatureValueLimitTotal2, config={"max_values_per_feature": 100000}
        )
        self.assertEqual(result.state.label, State.Label.COMPLETED_WITH_ERRORS)
        self.assertIn("too many values for plugin (30027) only returning", result.state.message)
        self.assertIn("dropping 508/10008 values from example_int", result.state.message)
        self.assertIn("dropping 10009/10009 values from example_string", result.state.message)
        self.assertIn("dropping 10010/10010 values from example_path", result.state.message)
        self.assertEqual(len(result.events[0].features["example_path"]), 0)
        self.assertEqual(len(result.events[0].features["example_int"]), 9500)
        self.assertEqual(len(result.events[0].features["example_string"]), 0)

    class DPReturnInfo(sup.DummyPlugin):
        def execute(self, job):
            self.add_info({"key": "value", "list": [1, 2, 5]})

    def test_return_info(self):
        """Tests that plugin can successfully return 'extra' values (unspecified structured data)"""

        result = self.do_execution(plugin_class=self.DPReturnInfo)
        self.assertEqual(result.state, State())
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", info={"key": "value", "list": [1, 2, 5]})],
            ),
        )

    class DPInvalidInfo(sup.DummyPlugin):
        def execute(self, job):
            self.add_info({"key": "value", "invalid": b"bytes"})

    def test_invalid_info(self):
        """Test that an error is returned when the 'extra' value isn't JSON-serialisable."""

        result = self.do_execution(plugin_class=self.DPInvalidInfo)
        print(result.state)
        self.assertEqual(result.state.label, State.Label.ERROR_OUTPUT)
        self.assertEqual(result.state.failure_name, "Invalid Plugin Output")
        self.assertIn("TypeError: Object of type bytes is not JSON serializable", result.state.message)
        self.assertRegex(result.state.message, "info was not JSON-serialisable")

    class DPEventPropertiesSet(sup.DummyPlugin):
        def execute(self, job):
            data = job.event.entity.datastreams[0]
            assert data.sha256 == "1669594220a92d73d62727293e988b4213b5b4829de36c3afe43c9b4f3ddf35e"
            assert data.size == 17
            assert data.file_format == "text/plain"
            assert data.magic == "ASCII text, with no line terminators"
            assert data.mime == "text/plain"
            assert job.event.entity.sha256 == "1669594220a92d73d62727293e988b4213b5b4829de36c3afe43c9b4f3ddf35e"
            assert job.event.entity.size == 17
            assert job.event.entity.file_format == "text/plain"

    def test_event_properties_set(self):
        """Test that an error is returned when the 'extra' value isn't JSON-serialisable."""

        result = self.do_execution(plugin_class=self.DPEventPropertiesSet, data_in=[("content", b"this is some text")])
        self.assertJobResult(result, JobResult(state=State(State.Label.COMPLETED_EMPTY)))

    class DPFeatIn(sup.DummyPlugin):
        def execute(self, job):
            f = job.event.entity.features[0]
            self.add_feature_values("example_string", f.value)

    def test_feats_in(self):
        """Test that feature values can be accessed.

        Important for plugins which use filepath for things, yara, dynamic, etc.
        """

        result = self.do_execution(
            plugin_class=self.DPFeatIn,
            feats_in=[azm.FeatureValue(name="filename", value="walrus.exe", type=azm.FeatureType.String)],
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"example_string": [FV("walrus.exe")]})],
            ),
        )

    class DPFeatureOrder(sup.DummyPlugin):
        FEATURES = [
            Feature("a", "demo", type=FeatureType.String),
            Feature("b", "demo", type=FeatureType.String),
            Feature("c", "demo", type=FeatureType.String),
            Feature("d", "demo", type=FeatureType.String),
            Feature("e", "demo", type=FeatureType.String),
            Feature("num", "demo", type=FeatureType.Integer),
        ]

        def execute(self, job):
            self.add_feature_values("e", "1")
            self.add_feature_values("c", "1")
            self.add_feature_values("a", "99")
            self.add_feature_values("b", "1")
            self.add_feature_values("d", "1")
            self.add_feature_values("a", "9")
            self.add_feature_values("a", "1")
            self.add_feature_values("a", "999")
            self.add_feature_values("num", 1)
            self.add_feature_values("num", 9)
            self.add_feature_values("num", 10)
            self.add_feature_values("num", 90)
            self.add_feature_values("num", 100)
            self.add_feature_values("num", 900)

    def test_feature_order(self):
        """Test that feature and feature value ordering is correct (re-sorted in assertJobResult)."""
        result = self.do_execution(plugin_class=self.DPFeatureOrder)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={
                            "b": [FV("1")],
                            "c": [FV("1")],
                            "d": [FV("1")],
                            "e": [FV("1")],
                            "a": [FV("1"), FV("9"), FV("99"), FV("999")],
                            "num": [FV(900), FV(1), FV(9), FV(10), FV(90), FV(100)],
                        },
                    )
                ],
            ),
        )

        self.assertEqual(["a", "b", "c", "d", "e", "num"], list(result.events[0].features.keys()))
        self.assertEqual([FV("1"), FV("9"), FV("99"), FV("999")], result.events[0].features["a"])

    class DPInspectData(sup.DummyPlugin):
        def execute(self, job):
            self.add_data("text", {"language": "numerals"}, b"12345")

    def test_inspect_data(self):
        """Test that data can be stripped."""

        result = self.do_execution(plugin_class=self.DPInspectData)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        data=[
                            EventData(
                                hash="5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5",
                                label=azm.DataLabel.TEXT,
                                language="numerals",
                            )
                        ],
                    )
                ],
                data={"5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5": b"12345"},
            ),
            inspect_data=True,
        )
        # test can still read data
        self.assertEqual(
            result.data["5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5"].read(), b"12345"
        )

    class DPStripData(sup.DummyPlugin):
        def execute(self, job):
            self.add_data("text", {"language": "numerals"}, b"12345")

    class DPStripData2(sup.DummyPlugin):
        def execute(self, job):
            self.add_child_with_data({"action": "barbeque"}, b"12345")

    def test_strip_data(self):
        """Test that data can be stripped."""

        result = self.do_execution(plugin_class=self.DPStripData)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        data=[
                            EventData(
                                hash="5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5",
                                label=azm.DataLabel.TEXT,
                                language="numerals",
                            )
                        ],
                    )
                ],
                data={"5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5": b""},
            ),
        )
        # test can still read data
        self.assertEqual(
            result.data["5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5"].read(), b"12345"
        )

        result = self.do_execution(plugin_class=self.DPStripData2)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5",
                        relationship={"action": "barbeque"},
                        data=[
                            EventData(
                                hash="5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5",
                                label=azm.DataLabel.CONTENT,
                            )
                        ],
                    )
                ],
                data={"5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5": b""},
            ),
        )
        self.assertEqual(
            result.data["5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5"].read(), b"12345"
        )

    class DPStripHash(sup.DummyPlugin):
        def execute(self, job):
            self.add_child_with_data({"action": "barbeque"}, b"12345")

    class DPStripHash2(sup.DummyPlugin):
        def execute(self, job):
            self.add_feature_values("example_string", "binary_bbq")
            self.add_child_with_data({"action": "barbeque"}, b"12345")
            self.add_data("text", {"language": "c++"}, b"123456")
            self.add_child_with_data({"action": "bbq3"}, b"centigrade")

    def test_strip_hash(self):
        """Test that hash can be stripped."""

        result = self.do_execution(plugin_class=self.DPStripHash)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="0",
                        relationship={"action": "barbeque"},
                        data=[EventData(hash="0", label=azm.DataLabel.CONTENT)],
                    )
                ],
                data={"0": b""},
            ),
            strip_hash=True,
        )

        result = self.do_execution(plugin_class=self.DPStripHash2)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        data=[EventData(hash="1", label=azm.DataLabel.TEXT, language="c++")],
                        features={"example_string": [FV("binary_bbq")]},
                    ),
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="0",
                        relationship={"action": "barbeque"},
                        data=[EventData(hash="0", label=azm.DataLabel.CONTENT)],
                    ),
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="2",
                        relationship={"action": "bbq3"},
                        data=[EventData(hash="2", label=azm.DataLabel.CONTENT)],
                    ),
                ],
                data={"0": b"", "1": b"", "2": b""},
            ),
            strip_hash=True,
        )

    class DPOldEventProperties(sup.DummyPlugin):
        def execute(self, job):
            self.add_feature_values("example_int", [4])

    def test_old_event_properties(self):
        result = self.do_execution(plugin_class=self.DPOldEventProperties)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        entity_id="test_entity",
                        entity_type="binary",
                        features={"example_int": [FV("4")]},
                    )
                ],
            ),
        )

    class DPAddSettings(sup.DummyPlugin):
        def execute(self, job: Job):
            test_val = job.event.source.settings.get(TEST_SETTING_KEY)
            self.add_feature_values("example_string", test_val)

    def test_add_settings(self):
        """
        Test adding settings to test case run.
        """

        # Check feature is wrapped and accepted if correct type

        result = self.do_execution(
            plugin_class=self.DPAddSettings,
            submission_settings={TEST_SETTING_KEY: "Random test setting value that seems fine."},
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        features={"example_string": [FV("Random test setting value that seems fine.")]},
                    )
                ],
            ),
        )



"""CLI based execution of plugins for local testing or remote deployment."""

import argparse
import base64
import contextlib
import datetime
import hashlib
import io
import json
import logging
import os
import struct
import sys
from io import BytesIO
from typing import Any, BinaryIO, Type

import cart
from azul_bedrock import identify
from azul_bedrock import models_network as azm
from pydantic import BaseModel

from azul_runner.log_setup import setup_logger

from . import local, settings
from .models import JobResult
from .monitor import Monitor
from .plugin import Plugin, State
from .storage import StorageProxyFile

logger = logging.getLogger(__name__)
MANDATORY_CART_HEADER_LEN = struct.calcsize(cart.MANDATORY_HEADER_FMT)


class Args(BaseModel):
    """Hold command line arguments."""

    config: list[tuple[str, str]] = []
    entity_id: str = ""
    files: str = ""
    job_limit: int | None = None
    output_folder: str = ""
    output_json: bool = False
    quiet: bool = False
    server: str = ""
    stream: Any = None
    verbose: int = 0


def cmdline_run(plugin: type[Plugin] = None):
    """Run from command-line."""
    if not plugin:
        print("Error: not run with a plugin", file=sys.stderr)
        exit(1)
    return execute(plugin, parse_args())


def args_to_config(args: Args) -> dict[str, Any]:
    """Convert input arguments into co-ordinator config settings."""
    config = {}
    if args.server:
        config["events_url"] = args.server
        config["data_url"] = args.server
    if args.config:
        # Update with `-c NAME VALUE` args
        config.update({n: v for n, v in args.config})
    return config


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Run an Azul plugin.")
    parser.add_argument(
        "-c",
        "--config",
        nargs=2,
        metavar=("NAME", "VALUE"),
        action="append",
        help="Provides config values for the plugin (and overrides values in config files). "
        "Only simple values are supported. Can be used multiple times.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Gives more verbose logging (pass once for info, twice for debug)",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Suppress warnings (log errors only). Overrides -v."
    )
    parser.add_argument(
        "--server",
        help="Fetch and execute jobs from the specified server. "
        "This is mutually exclusive with any of the local options below.",
        default=None,
    )
    parser.add_argument(
        "--job-limit",
        type=int,
        metavar="NUM",
        help="Exit after processing this many jobs. Only meaningful when connecting to a server.",
    )
    parser.add_argument(
        "-e",
        "--entity-id",
        default=None,
        help="Specify an entity-id for the entity to be processed. "
        "Most plugins should not require this to be explicitly set.",
    )
    parser.add_argument(
        "-o", "--output-json", action="store_true", help="Print JSON instead of human-readable summary."
    )
    parser.add_argument(
        "--output-folder",
        default=None,
        help="Write generated streams to this folder. Warning - Files are not neutered.",
    )
    parser.add_argument(
        "-s",
        "--stream",
        nargs=2,
        metavar=("TYPE", "FILENAME"),
        action="append",
        help="Read an input data stream of TYPE from the given FILENAME. Can be used multiple times.",
    )
    parser.add_argument(
        "files",
        nargs="?",
        help="Run plugin over file/folder recursively. Use - for stdin. Files will be loaded as 'content' streams.",
    )
    raw_args = parser.parse_args()

    # parse args
    setted = {}
    for k, v in raw_args.__dict__.items():
        if v is not None:
            setted[k] = v
    args = Args(**setted)

    # access log level from environment variable
    s = settings.RunnerSettings()

    # Log level settings
    if args.quiet:
        log_level = logging.ERROR
    elif args.verbose > 1:
        log_level = logging.DEBUG
    elif args.verbose > 0:
        log_level = logging.INFO
    else:
        # default to environment setting
        # logging library is incompatible between python versions 3.10 and 3.12 so this is being manually mapped.
        _nameToLevel = {
            "CRITICAL": logging.CRITICAL,
            "FATAL": logging.FATAL,
            "ERROR": logging.ERROR,
            "WARN": logging.WARNING,
            "WARNING": logging.WARNING,
            "INFO": logging.INFO,
            "DEBUG": logging.DEBUG,
            "NOTSET": logging.NOTSET,
        }
        # fall back to warning level
        log_level = _nameToLevel.get(s.log_level.upper(), logging.WARNING)

    # Setup Logger
    setup_logger(log_level)

    # print runtime options
    logger.info("command line run options:")
    for k, v in args.model_dump().items():
        logger.info(f"{k:20}: {v}")

    return args


def execute(plugin_class: type[Plugin], args: Args):
    """Execute the plugin with the provided arguments."""
    config = args_to_config(args)
    if not any((args.stream, args.files)):
        # FUTURE SIGINT handler for ^C to exit
        m_loop = Monitor(plugin_class, config)
        m_loop.run_loop()
    else:
        # local execution
        # Input / output JSON
        m_loop = Monitor(plugin_class, config)

        if not args.stream and not args.files:
            print("Error: No files/streams specified", file=sys.stderr)
            exit(1)

        if args.files and os.path.isdir(args.files):
            for path, _dir, files in os.walk(args.files):
                for file in sorted(files):
                    path_file = os.path.join(path, file)
                    print(path_file)
                    process_file(plugin_class, m_loop, path_file, args)
        else:
            process_file(plugin_class, m_loop, args.files, args)


def process_file(plugin_class: Type[Plugin], m_loop: Monitor, filepath, args: Args):
    """Run plugin once over streams."""
    sha256 = args.entity_id

    ent_sha512 = None
    ent_sha1 = None
    ent_md5 = None
    file_format = None
    file_extension = None
    size = None

    # Read files in
    spf_in: list[StorageProxyFile] = []
    fileinfo_in = []
    if filepath:
        if filepath == "-":
            # open stdin for reading
            data = sys.stdin.buffer.read()
        else:
            with open(filepath, "rb") as f:
                data = unpack_cart_and_read(f)
        if not sha256:
            sha256 = hashlib.sha256(data).hexdigest()
        ent_sha512 = hashlib.sha512(data).hexdigest()
        ent_sha1 = hashlib.sha1(data).hexdigest()  # noqa S324
        ent_md5 = hashlib.md5(data).hexdigest()  # noqa S324
        _, _, file_format, file_extension = identify.from_buffer(data)

        size = len(data)

        if len(data) == 0:
            logger.warning("Input content has zero length; did you mean to do this?")
        fileinfo = local.gen_api_content(io.BytesIO(data), label=azm.DataLabel.CONTENT)
        fileinfo_in.append(fileinfo)
        spf_in.append(
            StorageProxyFile(
                source="local",
                label=azm.DataLabel.CONTENT,
                hash="<content>",
                init_data=data,
                file_info=fileinfo,
                allow_unbounded_read=True,
            )
        )

    # Data streams
    for stype, sfile in args.stream or []:
        if stype == azm.DataLabel.CONTENT:
            raise Exception("use filepath for 'content' type, only one content binary is allowed")

        with open(sfile, "rb") as f:
            data = f.read()

        if len(data) == 0:
            logger.warning("Input stream (%s, %s) has zero length; did you mean to do this?" % (stype, sfile))
        fileinfo = local.gen_api_content(io.BytesIO(data), label=stype)
        fileinfo_in.append(fileinfo)
        spf_in.append(
            StorageProxyFile(
                source="local",
                label=stype,
                hash=sfile,
                init_data=data,
                file_info=fileinfo,
                allow_unbounded_read=True,
            )
        )

    # Use commandline provided file to calculate sha256 of file.
    if not sha256:
        sha256 = "cmdline_entity"

    # Validate input streams against plugin expectations
    try:
        local.validate_streams(spf_in, m_loop._plugin.cfg)
    except AssertionError as e:
        logger.warning("Input streams do not match plugin requirements (%s) - plugin may not run correctly" % e.args)

    entity = azm.BinaryEvent.Entity(
        size=size,
        datastreams=fileinfo_in,
        file_format=file_format,
        file_extension=file_extension,
        sha256=sha256,
        sha512=ent_sha512,
        sha1=ent_sha1,
        md5=ent_md5,
    )
    event = local.gen_event(entity)

    # Do execution
    res = m_loop.run_once(event, spf_in)

    [s.close() for s in spf_in if not s.closed]
    logger.info("Processing complete")
    if args.output_json:
        for result in res.values():
            # stick with jsonlines output in case of multiple results
            try:
                print(generate_json(result))
            except Exception as e:
                raise Exception(result) from e
    else:
        for subplugin, result in res.items():
            print_result(plugin_class, subplugin, result)

    # save plugin results to specific folder
    # not encoded, so raw malware may be written to disk
    if args.output_folder:
        try:
            os.mkdir(args.output_folder)
        except FileExistsError:
            pass

        # save all streams (may duplicate files)
        streams = []
        for jresult in res.values():
            for event in jresult.events:
                for edata in event.data:
                    streams.append((edata.label, edata.hash, jresult.data[edata.hash]))
        for stream in streams:
            with open(os.path.join(args.output_folder, f"{stream[1]}_{stream[0]}.data"), "wb") as f:
                f.write(stream[2].read())
                stream[2].seek(0)

    # Attempt to close and delete the file of origin for any open streams, this cleans up temp.
    streams = []
    for jresult in res.values():
        for data_stream in jresult.data.values():
            with contextlib.suppress(Exception):
                data_stream.close()
                os.remove(data_stream.name)


def generate_json(result: JobResult, indent=None):
    """Return the supplied JobResult as json string."""
    result = result.model_dump(exclude_defaults=True)

    # Dump the JSON
    def object_to_json(obj):
        if isinstance(obj, set):
            return sorted(list(obj))
        elif isinstance(obj, bytes):
            return base64.b64encode(obj).decode("ascii")
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        else:
            raise TypeError(f"Unimplemented JSON encode called for {type(obj)} object: \n{obj}")

    # stick with jsonlines output in case of multiple results
    return json.dumps(result, default=object_to_json, indent=indent)


def print_result(plugin_class: Type[Plugin], subplugin: str, result: JobResult):
    """Print result of plugin to stdout."""
    # only mention multiplugin if one is present
    if subplugin and subplugin is not None:
        print("----- %s-%s results -----" % (plugin_class.__name__, subplugin))
    else:
        print("----- %s results -----" % plugin_class.__name__)

    if getattr(plugin_class, "SECURITY", False):
        print("** SECURITY: %s **" % plugin_class.SECURITY)
    status: State = result.state
    if status.failure_name:
        print("%s (%s)" % (status.label.name, status.failure_name))
    else:
        print(status.label.name)
    if status.message:
        print(status.message)
    print()

    if result.events:
        print(f"events ({len(result.events)})")
        print()
        for event in result.events:
            # print event info basic
            if not event.sha256 and not event.parent_sha256:
                print("event for primary")
            elif event.parent_sha256:
                print(f"event for data {event.parent_sha256}")
            else:
                print(f"event for {event.sha256}:{event.parent_sha256}")
            print(f"  {event.relationship}")
            if (p := event.parent) is not None:
                if not p.sha256:
                    print("")
                else:
                    print(f"  child of {p.sha256}")
            # print child info
            if event.data:
                print(f"  output data streams ({len(event.data)}):")
                for data in event.data:
                    bin = result.data[data.hash]
                    # get size of result file
                    bin.seek(0, 2)
                    size = bin.tell()
                    bin.seek(0)
                    print(f"    {size} bytes - {data}")
                    bin.seek(0)
            if event.features:
                print("  output features:")
                fmt = "    %" + str(max([len(f) for f in event.features])) + "s: %s"
                fmt2 = " " * max([len(f) for f in event.features]) + "      %s"
                for fn, fvs in event.features.items():
                    for i, fv in enumerate(sorted(fvs)):
                        # Build value sting: <label> - <value> @ <offset> (offset)
                        values_str = f"{fv.value}"
                        if fv.label:
                            values_str = f"{fv.label} - " + values_str
                        if fv.offset:
                            values_str = values_str + f" @ {hex(fv.offset)} (offset)"

                        if not i:
                            print(fmt % (fn, values_str))
                        else:
                            print(fmt2 % values_str)
            if event.info:
                print("  info:")
                for k, v in event.info.items():
                    print("    %s: %s" % (k, v))
            print()

    if result.feature_types:
        print("Feature key:")
        for f in result.feature_types:
            print("  %s:  %s" % (f.name, f.desc))
        print()


def unpack_cart_and_read(input_stream: BinaryIO) -> bytes:
    """Conditionally unpacks CaRT files and reads bytes from provided filepath."""
    # Ensure read from start of the file
    input_stream.seek(0)
    header = input_stream.read(MANDATORY_CART_HEADER_LEN)
    # Reset the read head, incase it is not a CaRT file, the function can just return
    input_stream.seek(0)
    if cart.is_cart(header):
        unpacked = BytesIO()
        cart.unpack_stream(input_stream, unpacked)
        unpacked.seek(0)
        return unpacked.getvalue()
    else:
        return input_stream.read()


if __name__ == "__main__":
    cmdline_run()



import datetime
import hashlib
import io
import json
import unittest
from typing import Any, Dict
from unittest import mock

from azul_bedrock import models_network as azm
from pydantic import TypeAdapter

from azul_runner import (
    FV,
    Event,
    EventData,
    EventParent,
    Feature,
    FeatureType,
    FeatureValue,
    Job,
    JobResult,
    Plugin,
    State,
    StorageProxyFile,
    TestPlugin,
    add_settings,
    coordinator,
    local,
    network,
    network_transform,
    settings,
)
from azul_runner.plugin import Multiplugin


class DP(Plugin):
    SETTINGS = add_settings(
        request_retry_count=0,
        server="https://localhost:9876",
        filter_data_types={},
    )  # Don't retry failed requests when testing

    # leave security property unset
    # SECURITY = None
    VERSION = "1.0"
    MULTI_STREAM_AWARE = True
    FEATURES = [
        Feature("apples", "Example feature", type=FeatureType.Integer),
    ]

    def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
        super().__init__(config)
        self.register_multiplugin(
            "decode_rot", "1.0.0", self.decode_rot, description="Description decode_rot", security="LEMON RUSH"
        )
        self.register_multiplugin("decode_xor", "21-02-2025", self.decode_xor, description="Description decode_xor")
        self.register_multiplugin("decode_nor", "v1-beta", self.decode_nor)

    def execute(self, job) -> dict:
        return

    def decode_rot(self, job) -> dict:
        return

    def decode_xor(self, job) -> dict:
        return

    def decode_nor(self, job) -> dict:
        return


class TestBasic(unittest.TestCase):
    def test_basic(self):
        class DummyPlugin(DP):
            """Dummy plugin description."""

            def decode_rot(self, job):
                self.add_feature_values("apples", [FV(1)])

            def decode_xor(self, job):
                raise Exception("invalid")

            def decode_nor(self, job):
                c = self._add_child("1", {"a": 1})
                c.add_data("content", {}, b"5")
                self.add_info({"k": "1234567890"})

        loop = coordinator.Coordinator(DummyPlugin, settings.parse_config(DummyPlugin, {}))

        # test registration events
        self.assertEqual(len(network_transform.get_registrations(loop._plugin)), 4)
        self.assertEqual(network_transform.get_registrations(loop._plugin)[0].author.name, "DummyPlugin")
        self.assertEqual(network_transform.get_registrations(loop._plugin)[1].author.name, "DummyPlugin-decode_rot")
        self.assertEqual(network_transform.get_registrations(loop._plugin)[2].author.name, "DummyPlugin-decode_xor")
        self.assertEqual(network_transform.get_registrations(loop._plugin)[3].author.name, "DummyPlugin-decode_nor")

        self.assertEqual(
            network_transform.get_registrations(loop._plugin)[0].entity.description, "Dummy plugin description."
        )
        self.assertEqual(
            network_transform.get_registrations(loop._plugin)[1].entity.description, "Description decode_rot"
        )
        self.assertEqual(
            network_transform.get_registrations(loop._plugin)[2].entity.description, "Description decode_xor"
        )
        self.assertEqual(
            network_transform.get_registrations(loop._plugin)[3].entity.description, "Dummy plugin description."
        )

        self.assertSetEqual(
            set(loop._plugin.get_registered_multiplugins()), set(["decode_rot", "decode_xor", "decode_nor", None])
        )
        self.assertEqual(
            loop._plugin.get_multiplugin("decode_rot"),
            Multiplugin(
                name="decode_rot",
                version="1.0.0",
                callback=loop._plugin.decode_rot,
                description="Description decode_rot",
                security="LEMON RUSH",
            ),
        )

        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        results = loop.run_once(local.gen_event(entity))
        self.assertEqual(set(results.keys()), {None, "decode_rot", "decode_xor", "decode_nor"})

        sub = results["decode_rot"]
        sub.runtime = None
        sub.date_start = None
        sub.date_end = None
        self.assertEqual(sub.state, State())
        self.assertEqual(
            sub.feature_types, [Feature(name="apples", desc="Example feature", type=azm.FeatureType.Integer)]
        )

        self.assertEqual(
            TypeAdapter(list[Event]).dump_python(sub.events),
            TypeAdapter(list[Event]).dump_python(
                [
                    Event(
                        sha256="id",
                        features={"apples": {FeatureValue(value=1, offset=None, size=None)}},
                    )
                ]
            ),
        )

        sub = results["decode_xor"]
        sub.runtime = None
        sub.date_start = None
        sub.date_end = None
        self.assertEqual(sub.state.label, State.Label.ERROR_EXCEPTION)

        sub = results["decode_nor"]
        sub.runtime = None
        sub.date_start = None
        sub.date_end = None
        self.assertEqual(sub.state, State())
        self.assertEqual(sub.feature_types, [])
        self.assertIn("ef2d127de37b942baad06145e54b0c619a1f22327b2ebbcfbec78f5564afe39d", sub.data)
        self.assertEqual(
            TypeAdapter(list[Event]).dump_python(sub.events),
            TypeAdapter(list[Event]).dump_python(
                [
                    Event(sha256="id", info={"k": "1234567890"}),
                    Event(
                        parent=EventParent(sha256="id"),
                        sha256="1",
                        data=[
                            {
                                "hash": "ef2d127de37b942baad06145e54b0c619a1f22327b2ebbcfbec78f5564afe39d",
                                "label": "content",
                            }
                        ],
                        relationship={"a": 1},
                    ),
                ]
            ),
        )


class TestRegistration(unittest.TestCase):
    @mock.patch("azul_bedrock.dispatcher.DispatcherAPI.submit_events")
    def test_registration(self, reqw):
        class DummyPlugin(DP):
            pass

        p = DummyPlugin()
        net = network.Network(p)
        net.post_registrations()

        author_events: list[azm.PluginEvent] = [x[0][0][0] for x in reqw.call_args_list]

        self.assertEqual(len(author_events), 4)

        self.assertEqual(author_events[0].author, azm.Author(category="plugin", version="1.0", name="DummyPlugin"))
        self.assertEqual(
            author_events[1].author,
            azm.Author(category="plugin", version="1.0-1.0.0", name="DummyPlugin-decode_rot", security="LEMON RUSH"),
        )
        self.assertEqual(
            author_events[2].author,
            azm.Author(category="plugin", version="1.0-21-02-2025", name="DummyPlugin-decode_xor"),
        )
        self.assertEqual(
            author_events[3].author,
            azm.Author(category="plugin", version="1.0-v1-beta", name="DummyPlugin-decode_nor"),
        )

    @mock.patch("azul_bedrock.dispatcher.DispatcherAPI.submit_events")
    def test_registration_no_version(self, reqw):
        class DummyPlugin(DP):
            def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
                super().__init__(config)
                self._multiplugins = {}  # remove parent plugins
                self.register_multiplugin(None, None, lambda x: None)
                self.register_multiplugin("DecodeRot", "", self.decode_rot, description="Custom description here")

        p = DummyPlugin()
        net = network.Network(p)
        net.post_registrations()
        author_events: list[azm.PluginEvent] = [x[0][0][0] for x in reqw.call_args_list]
        self.assertEqual(len(author_events), 2)

        self.assertEqual(author_events[0].author, azm.Author(category="plugin", version="1.0", name="DummyPlugin"))
        self.assertEqual(
            author_events[1].author, azm.Author(category="plugin", version="1.0", name="DummyPlugin-DecodeRot")
        )
        self.assertEqual(author_events[1].entity.description, "Custom description here")


class TestLoop(unittest.TestCase):
    def setUp(self) -> None:
        class DummyPlugin(DP):
            def decode_rot(self, job):
                self.add_feature_values("apples", [FV(1)])

            def decode_xor(self, job):
                raise Exception("invalid")

            def decode_nor(self, job):
                c = self._add_child("1", {"a": 1})
                c.add_data("content", {}, b"5")
                c.add_feature_values("apples", [FV(1)])
                self.add_info({"k": "1234567890"})

        self.ploop = coordinator.Coordinator(DummyPlugin, settings.parse_config(DummyPlugin, {}))
        content = b"small content"
        content_sha = hashlib.sha256(content).hexdigest()
        self.streams = [
            StorageProxyFile(
                source="local",
                label="",
                hash="",
                init_data=content,
                file_info=local.gen_api_content(io.BytesIO(content)),
            )
        ]
        self.event = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="gregre",
            action=azm.BinaryAction.Enriched,
            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            source=azm.Source(
                name="source",
                path=[],
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            ),
            author=azm.Author(name="TestServer", category="blah"),
            entity=azm.BinaryEvent.Entity(
                sha256="1234",
                datastreams=[
                    azm.Datastream(
                        identify_version=1,
                        label=azm.DataLabel.CONTENT,
                        size=13,
                        sha256=content_sha,
                        sha1="1",
                        md5="5",
                        sha512="512",
                        mime="mt",
                        magic="mm",
                        file_format="ft",
                    ),
                ],
            ),
        )

        self.job = Job(event=self.event)
        self.job.load_streams(local=self.streams)
        return super().setUp()

    def test_run_once_multiplugin(self):
        """Test that individual multiplugins can be run for event."""
        resp_result = self.ploop._run_job_with_multiplugin(self.job, None)
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED_EMPTY)

        resp_result = self.ploop._run_job_with_multiplugin(self.job, "decode_rot")
        self.assertEqual(resp_result.events[0].features, {"apples": [FV(1)]})
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED)

        resp_result = self.ploop._run_job_with_multiplugin(self.job, "decode_xor")
        self.assertEqual(resp_result.state.label, State.Label.ERROR_EXCEPTION)

        resp_result = self.ploop._run_job_with_multiplugin(self.job, "decode_nor")
        self.assertEqual(
            TypeAdapter(list[Event]).dump_python(resp_result.events),
            TypeAdapter(list[Event]).dump_python(
                [
                    Event(sha256="1234", info={"k": "1234567890"}),
                    Event(
                        parent=EventParent(sha256="1234"),
                        sha256="1",
                        data=[
                            {
                                "hash": "ef2d127de37b942baad06145e54b0c619a1f22327b2ebbcfbec78f5564afe39d",
                                "label": "content",
                            }
                        ],
                        relationship={"a": 1},
                        features={"apples": {FeatureValue(value=1, offset=None, size=None)}},
                    ),
                ]
            ),
        )
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED)

    def test_run_once_ack(self):
        calls = [x for x in self.ploop._run_job(self.job.event)]
        self.assertEqual(len(calls), 4)

        resp_result = calls[0][0]
        self.assertEqual(resp_result.events[0].features, {"apples": [FV(1)]})
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED)
        multiplugin = calls[0][1]
        self.assertEqual(multiplugin, "decode_rot")

        resp_result = calls[1][0]
        self.assertEqual(resp_result.state.label, State.Label.ERROR_EXCEPTION)
        multiplugin = calls[1][1]
        self.assertEqual(multiplugin, "decode_xor")

        resp_result = calls[2][0]
        self.assertEqual(
            TypeAdapter(list[Event]).dump_python(resp_result.events),
            TypeAdapter(list[Event]).dump_python(
                [
                    Event(sha256="1234", info={"k": "1234567890"}),
                    Event(
                        parent=EventParent(
                            sha256="1234",
                        ),
                        sha256="1",
                        data=[
                            {
                                "hash": "ef2d127de37b942baad06145e54b0c619a1f22327b2ebbcfbec78f5564afe39d",
                                "label": "content",
                            }
                        ],
                        relationship={"a": 1},
                        features={"apples": {FeatureValue(value=1, offset=None, size=None)}},
                    ),
                ]
            ),
        )
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED)
        multiplugin = calls[2][1]
        self.assertEqual(multiplugin, "decode_nor")

        resp_result = calls[3][0]
        self.assertFalse(resp_result.events)
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED_EMPTY)

    def test_run_once_ack_optout(self):
        class DummyPlugin(DP):
            def execute(self, job):
                return State.Label.OPT_OUT

        self.ploop = coordinator.Coordinator(DummyPlugin, settings.parse_config(DummyPlugin, {}))

        calls = [x for x in self.ploop._run_job(self.job.event)]
        self.assertEqual(len(calls), 1)

        resp_result = calls[0][0]
        self.assertNotIn("features", resp_result)
        self.assertEqual(resp_result.state.label, State.Label.OPT_OUT)

    def test_run_completed_with_errors(self):
        class DummyPlugin(DP):
            def execute(self, job):
                return State.Label.COMPLETED_WITH_ERRORS

        self.ploop = coordinator.Coordinator(DummyPlugin, settings.parse_config(DummyPlugin, {}))

        calls = [x for x in self.ploop._run_job(self.job.event)]
        self.assertEqual(len(calls), 4)
        # Sub-plugins completed with no data
        resp_result = calls[0][0]
        self.assertNotIn("features", resp_result)
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED_EMPTY)

        # Parent plugin completed with errors.
        resp_result = calls[-1][0]
        self.assertNotIn("features", resp_result)
        self.assertEqual(resp_result.state.label, State.Label.COMPLETED_WITH_ERRORS)


class TestPluginExecute(TestPlugin):
    """Tests for the handling of add_entity_content."""

    class DummyPlugin(DP):
        def decode_rot(self, job):
            self.add_feature_values("apples", [FV(1)])

        def decode_xor(self, job):
            raise Exception("invalid")

        def decode_nor(self, job):
            c = self._add_child("1", {"a": 1})
            c.add_data("content", {}, b"5")
            c.add_feature_values("apples", [FV(1)])
            self.add_info({"k": "1234567890"})

    PLUGIN_TO_TEST = DummyPlugin

    def test_multi_output(self):
        """Test that multiplugin results are passed through to test cases."""
        result = self.do_execution()

        self.assertEqual(set(result.keys()), {None, "decode_rot", "decode_xor", "decode_nor"})
        self.assertJobResult(
            result["decode_rot"],
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"apples": [FV(1)]})],
            ),
            inspect_data=True,
        )
        self.assertEqual(result["decode_xor"].state.label, State.Label.ERROR_EXCEPTION)
        self.assertJobResult(
            result["decode_nor"],
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(sha256="test_entity", info={"k": "1234567890"}),
                    Event(
                        parent=EventParent(sha256="test_entity"),
                        sha256="1",
                        relationship={"a": 1},
                        data=[
                            EventData(
                                hash="ef2d127de37b942baad06145e54b0c619a1f22327b2ebbcfbec78f5564afe39d",
                                label=azm.DataLabel.CONTENT,
                            )
                        ],
                        features={"apples": [FV(1)]},
                    ),
                ],
                data={"ef2d127de37b942baad06145e54b0c619a1f22327b2ebbcfbec78f5564afe39d": b"5"},
            ),
            inspect_data=True,
        )


class TestPluginExecuteOptout(TestPlugin):
    """
    Tests for the handling of add_entity_content
    """

    class DummyPlugin(DP):
        def execute(self, job):
            return State.Label.OPT_OUT

        def decode_rot(self, job):
            self.add_feature_values("apples", [FV(1)])

        def decode_xor(self, job):
            raise Exception("invalid")

        def decode_nor(self, job):
            self._add_child("1", {"a": 1}, child_data=[("content", b"5")], child_features={"apples": [FV(1)]})
            self.add_info({"k": "1234567890"})

    PLUGIN_TO_TEST = DummyPlugin

    def test_multi_output(self):
        """Test that multiplugin results are passed through to test cases."""
        result = self.do_execution()
        # check that only one result was returned
        self.assertEqual(result.state.label, State.Label.OPT_OUT)



from __future__ import annotations

from azul_runner import (
    BinaryPlugin,
    Feature,
    FeatureType,
    FeatureValue,
    Job,
    State,
    TestPlugin,
    add_settings,
)


class BinaryTestDatalessPlugin(BinaryPlugin):
    VERSION = "1.0"
    FEATURES = [
        Feature("sample_feature", "An output feature for the test plugin", type=FeatureType.String),
    ]

    SETTINGS = add_settings(
        assume_streams_available=False,
        # any data types
        filter_data_types={},
    )

    def execute(self, job: Job) -> dict:
        digest = job.event.entity.sha512
        self.add_feature_values("sample_feature", digest)


class TestBinaryTemplateDataless(TestPlugin):
    PLUGIN_TO_TEST = BinaryTestDatalessPlugin

    def test_simple_run(self):
        """Check that the dataless test plugin runs and returns a basic feature."""

        result = self.do_execution(entity_attrs={"sha512": "a" * 64})
        self.assertEqual(result.state, State())
        self.assertEqual(result.main.features, {"sample_feature": [FeatureValue("a" * 64)]})



import unittest

from azul_runner import FV, Event, Feature, FeatureType, JobResult, State


class TestGenEvents(unittest.TestCase):
    def test_feature(self):
        self.assertTrue(Feature(name="f1", desc="", type=FeatureType.Integer))
        self.assertTrue(Feature(name="f1", desc="", type=FeatureType.Float))
        self.assertTrue(Feature(name="f1", desc="", type=FeatureType.String))
        self.assertTrue(Feature(name="f1", desc="", type=FeatureType.Binary))
        self.assertTrue(Feature(name="f1", desc="", type=FeatureType.Datetime))
        self.assertTrue(Feature(name="f1", desc="", type=FeatureType.Filepath))
        self.assertTrue(Feature(name="f1", desc="", type=FeatureType.Uri))
        self.assertRaises(ValueError, Feature, *("f1", "", dict))

    def test_event(self):
        # check ordinary JobResult also sorts automatically
        tmp = JobResult(
            state=State(State.Label.COMPLETED),
            events=[
                Event(
                    sha256="test_entity",
                    features={
                        "b": [FV("1")],
                        "c": [FV("1")],
                        "d": [FV("1")],
                        "e": [FV("1")],
                        "a": [FV("999"), FV("1"), FV("9"), FV("99")],
                    },
                )
            ],
        )
        self.assertEqual(["a", "b", "c", "d", "e"], list(tmp.events[0].features.keys()))
        self.assertEqual([FV("1"), FV("9"), FV("99"), FV("999")], tmp.events[0].features["a"])



"""Handle full Plugin execution when the plugin sources the original event itself rather than dispatcher.

* Utility to Generate and send sourced events to an external source.
* Run job based on an input event.
* Send job results to external source
* Run required multiplugins
* Handle timeouts and critical errors
"""

import hashlib
import io
import logging
from typing import Any, Type

import pendulum
from azul_bedrock import models_network as azm

from azul_runner import settings, storage
from azul_runner.coordinator import Coordinator
from azul_runner.network_transform import gen_author
from azul_runner.plugin import Plugin

logger = logging.getLogger(__name__)


def md5(text: str):
    """Return string md5 representing incoming text."""
    return hashlib.md5(text.encode()).hexdigest()  # noqa: S303, S324


class Pusher(Coordinator):
    """Allows plugins to publish binary sourced events with some basic structure.

    Note: not thread safe because of the way plugin features are tracked.
    """

    def __init__(self, plugin_class: Type[Plugin], config: dict[str, Any], is_register_plugin: bool = True):
        """Start the pusher with the provided configuration."""
        super().__init__(plugin_class, settings.parse_config(plugin_class, config))
        if not self.plugin._IS_USING_PUSHER:
            raise ValueError("Pusher cannot be used unless the _IS_USING_PUSHER flag is set to true.")

        if is_register_plugin:
            # Register the plugin.
            self._network.post_registrations()

    def generate_base_mapped_source(self, source_label: str, references: dict[str, str], security: str) -> azm.Source:
        """Generate a base Source event that can be used for pushing depth 0 (root level) mapped events."""
        timestamp = pendulum.now(pendulum.UTC)
        return azm.Source(
            name=source_label,
            timestamp=timestamp,
            references=references,
            path=[],
            security=security,
        )

    def _gen_mapped_event(
        self,
        *,
        source_info: azm.Source,
        source_file_info: azm.FileInfo,
        security: str,
        relationship: dict[str, str] | None = None,
        filename: str = "",
    ) -> azm.BinaryEvent:
        """Generate mapped events based on the existing source and file information."""
        sha256 = source_file_info.sha256
        if not sha256:
            raise Exception(f"Source file info must have its sha256 set and it is currently {sha256}")

        input_entity = source_file_info.to_input_entity()
        if filename:
            input_entity.features.append(
                azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value=filename)
            )
        # If this is a root level event use the same timestamp as the source event.
        if len(source_info.path) == 0:
            timestamp = pendulum.from_timestamp(source_info.timestamp.timestamp())
        else:
            timestamp = pendulum.now(pendulum.UTC)

        if not relationship:
            relationship = {}

        author = gen_author(self.plugin, None)
        author.security = security

        source_copy = source_info.model_copy(deep=True)
        source_copy.path.append(
            azm.PathNode(
                sha256=sha256,
                action=azm.BinaryAction.Mapped,
                timestamp=timestamp,
                author=author,
                relationship=relationship,
                file_format=input_entity.file_format,
                size=input_entity.size,
                filename=filename if filename else None,
            )
        )

        return azm.BinaryEvent(
            kafka_key=f"{author.name}-placeholder",  # temporary id so we can create the object
            dequeued=".".join(
                [
                    sha256,
                    author.name,
                    author.version if author.version else "",
                    timestamp.to_iso8601_string(),
                ]
            ),
            action=azm.BinaryAction.Mapped,
            model_version=azm.CURRENT_MODEL_VERSION,
            timestamp=timestamp,
            author=author,
            entity=input_entity,
            source=source_copy,
        )

    def _gen_sourced_event(
        self,
        *,
        input_entity: azm.BinaryEvent.Entity,
        source_label: str,
        references: dict[str, str],
        filename: str = "",
        security: str | None = None,
    ) -> azm.BinaryEvent:
        """Generate a sourced event that can have features added to it."""
        timestamp = pendulum.now(pendulum.UTC)
        if not input_entity.sha256:
            raise ValueError(f"Input entity must have it's sha256 set and it is currently {input_entity.sha256}")

        author = gen_author(self.plugin, None)
        author.security = security

        return azm.BinaryEvent(
            kafka_key=f"{author.name}-placeholder",  # temporary id so we can create the object
            dequeued=".".join(
                [
                    input_entity.sha256,
                    author.name,
                    author.version if author.version else "",
                    timestamp.to_iso8601_string(),
                ]
            ),
            action=azm.BinaryAction.Sourced,
            model_version=azm.CURRENT_MODEL_VERSION,
            timestamp=timestamp,
            author=author,
            entity=input_entity,
            source=azm.Source(
                name=source_label,
                timestamp=timestamp,
                references=references,
                path=[
                    azm.PathNode(
                        author=author,
                        action=azm.BinaryAction.Sourced,
                        timestamp=timestamp,
                        sha256=input_entity.sha256,
                        filename=filename if filename else None,
                        size=input_entity.size,
                        file_format=input_entity.file_format,
                    )
                ],
                security=security,
            ),
        )

    def push_once_sourced(
        self,
        *,
        content: bytes,
        source_label: str,
        references: dict[str, str],
        security: str,
        filename: str = "",
        local: list[storage.StorageProxyFile] | None = None,
    ) -> str:
        """Run the plugin on the provided input content and generate a sourced event for the provided source_label.

        Returns the sha256 of the provided content.
        """
        sha256 = storage.calc_stream_hash(io.BytesIO(content), hashlib.sha256)

        logger.info(f"Uploading - {sha256} - {filename} - {max(len(content) // 1024, 1)}kb - {security}")

        file_info_dict = self._network._post_data(
            source_label,
            {sha256: ([azm.DataLabel.CONTENT], io.BytesIO(content))},
        )
        file_info = file_info_dict[sha256]

        if file_info is None:
            raise Exception(f"The file with sha256 {sha256} was uploaded but no metadata was returned.")
        if file_info.sha256 is None:
            file_info.sha256 = sha256

        input_entity = file_info.to_input_entity()

        if filename:
            input_entity.features.append(
                azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value=filename)
            )
        # Sourced event because content was provied.
        root_event = self._gen_sourced_event(
            input_entity=input_entity,
            source_label=source_label,
            references=references,
            filename=filename,
            security=security,
        )
        self._run_plugin(root_event, local=local)
        return sha256

    def push_once_mapped(
        self,
        *,
        source_file_info: azm.FileInfo,
        source_info: azm.Source,
        security: str,
        relationship: dict[str, str] | None = None,
        filename: str = "",
        local: list[storage.StorageProxyFile] | None = None,
    ):
        """Map the provided metadata onto the provided source_info as a mapped event.

        NOTE - if this is a newly generated event use the generate_base_mapped_source function on pusher.
        """
        root_event = self._gen_mapped_event(
            source_file_info=source_file_info,
            security=security,
            source_info=source_info,
            relationship=relationship,
            filename=filename,
        )
        return self._run_plugin(root_event, local=local)

    def _run_plugin(
        self,
        root_event: azm.BinaryEvent,
        local: list[storage.StorageProxyFile] | None = None,
    ):
        if not self._cfg.events_url:
            raise ValueError("Cannot run and post results when events_url is None")

        if self._watched and self._watched.check_updated():
            self._recreate_plugin()
            logger.info("Plugin has been restarted due to a file change")

        try:
            # NOTE - root_event can be a mapped or sourced event depending on if a content stream was provided.
            for _result, _multi in self._run_job(root_event, local=local):
                self._network.ack_job(root_event, _result, _multi)
                # close any open handles in JobResult
                _result.close()

        except Exception as e:  # Unknown error so extra logging required.
            logger.error(
                f"Error {e} during job with kafka_key: '{root_event.kafka_key}', "
                + f"action: '{root_event.action}', author info: {root_event.author.model_dump()}"
                + f"\n source info: {root_event.source.model_dump()}."
            )
            raise
